@@ -1,4 +1,3 @@
-# -*- coding: UTF-8 -*-
 # This is a collection of "tests" done on the package data. The resut of the
 # tests is used to give the package a rating.
 #
@@ -15,30 +14,56 @@
 #     test(data): Performs the test on the given data. Returns True for pass
 #                 False for fail and None for not applicable (meaning it will
 #                 not be counted).
-
+import io
 import re
+from collections import defaultdict
+
 from docutils.core import publish_parts
 from docutils.utils import SystemMessage
+from trove_classifiers import classifiers as CLASSIFIERS
+from packaging.specifiers import InvalidSpecifier, SpecifierSet
 
-try:
-    stringtypes = basestring,
-except NameError:
-    stringtypes = str,
+LEVELS = [
+    "This cheese seems to contain no dairy products",
+    "Vieux Bologne",
+    "Limburger",
+    "Gorgonzola",
+    "Stilton",
+    "Brie",
+    "Comté",
+    "Jarlsberg",
+    "Philadelphia",
+    "Cottage Cheese",
+    "Your cheese is so fresh most people think it's a cream: Mascarpone",
+]
 
-LEVELS = [u"This cheese seems to contain no dairy products",
-          u"Vieux Bologne",
-          u"Limburger",
-          u"Gorgonzola",
-          u"Stilton",
-          u"Brie",
-          u"Comté",
-          u"Jarlsberg",
-          u"Philadelphia",
-          u"Cottage Cheese",
-          u"Your cheese is so fresh most people think it's a cream: Mascarpone"]
+SHORT_NAME_RE = re.compile(r"\(.*?\)")
 
 
-class BaseTest(object):
+def get_code_licenses():
+    licenses = [each for each in list(CLASSIFIERS) if each.startswith("License")]
+    code_map = defaultdict(set)
+    for license in licenses:
+        short_name = SHORT_NAME_RE.findall(license)
+        if short_name:
+            short_name = short_name[0][1:-1]
+            code_map[short_name].add(license)
+            if short_name.startswith("GPL"):
+                code_map["GPL"].add(license)
+            elif short_name.startswith("LGPL"):
+                code_map["LGPL"].add(license)
+        elif "Zope" in license:
+            code_map["ZPL"].add(license)
+        elif "MIT License" in license:
+            code_map["MIT"].add(license)
+
+    return code_map
+
+
+CODE_LICENSES = get_code_licenses()
+
+
+class BaseTest:
     fatal = False
 
 
@@ -49,18 +74,17 @@ class FieldTest(BaseTest):
         return bool(data.get(self.field))
 
     def message(self):
-        return ("Your package does not have %s data" % self.field) + (
-            self.fatal and '!' or '.')
+        return (f"Your package does not have {self.field} data") + (self.fatal and "!" or ".")
 
 
 class Name(FieldTest):
     fatal = True
-    field = 'name'
+    field = "name"
 
 
 class Version(FieldTest):
     fatal = True
-    field = 'version'
+    field = "version"
 
 
 class VersionIsString(BaseTest):
@@ -68,13 +92,15 @@ class VersionIsString(BaseTest):
 
     def test(self, data):
         # Check that the version is a string
-        version = data.get('version')
-        return isinstance(version, stringtypes)
+        version = data.get("version")
+        return isinstance(version, str)
 
     def message(self):
-        return 'The version number should be a string.'
+        return "The version number should be a string."
 
-PEP386_RE = re.compile(r'''
+
+PEP386_RE = re.compile(
+    r"""
     ^
     (?P<version>\d+\.\d+)          # minimum 'N.N'
     (?P<extraversion>(?:\.\d+)*)   # any number of extra '.N' segments
@@ -84,10 +110,13 @@ PEP386_RE = re.compile(r'''
         (?P<prerelversion>\d+(?:\.\d+)*)
     )?
     (?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?
-    $''', re.VERBOSE | re.IGNORECASE)
+    $""",
+    re.VERBOSE | re.IGNORECASE,
+)
 
 
-PEP440_RE = re.compile(r"""^
+PEP440_RE = re.compile(
+    r"""^
     v?
     (?:
         (?:(?P<epoch>[0-9]+)!)?                           # epoch
@@ -116,7 +145,10 @@ PEP440_RE = re.compile(r"""^
         )?
     )
     (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
-$""", re.VERBOSE | re.IGNORECASE)
+$""",
+    re.VERBOSE | re.IGNORECASE,
+)
+
 
 class PEPVersion(BaseTest):
     weight = 50
@@ -124,7 +156,7 @@ class PEPVersion(BaseTest):
 
     def test(self, data):
         # Check that the version number complies to PEP-386:
-        version = data.get('version')
+        version = data.get("version")
         self.pep386 = False
         if PEP386_RE.search(str(version)) is not None:
             # Matches the old PEP386
@@ -143,7 +175,7 @@ class Description(BaseTest):
     weight = 100
 
     def test(self, data):
-        description = data.get('description')
+        description = data.get("description")
         if not description:
             # No description at all. That's fatal.
             self.fatal = True
@@ -153,19 +185,18 @@ class Description(BaseTest):
 
     def message(self):
         if self.fatal:
-            return 'The package had no description!'
+            return "The package had no description!"
         else:
-            return ("The package's description should be longer than "
-                    "10 characters.")
+            return "The package's description should be longer than 10 characters."
 
 
 class LongDescription(BaseTest):
     weight = 50
 
     def test(self, data):
-        long_description = data.get('long_description', '')
-        if not isinstance(long_description, stringtypes):
-            long_description = ''
+        long_description = data.get("long_description", "")
+        if not isinstance(long_description, str):
+            long_description = ""
         return len(long_description) > 100
 
     def message(self):
@@ -174,18 +205,35 @@ class LongDescription(BaseTest):
 
 class Classifiers(FieldTest):
     weight = 100
-    field = 'classifiers'
+    field = "classifiers"
 
 
-class PythonVersion(BaseTest):
+class ClassifierVerification(BaseTest):
+    weight = 20
 
+    def test(self, data):
+        self._incorrect = []
+        classifiers = data.get("classifiers", [])
+        for classifier in classifiers:
+            if classifier not in CLASSIFIERS:
+                self._incorrect.append(classifier)
+        if self._incorrect:
+            return False
+        return True
+
+    def message(self):
+        err = "\n".join(self._incorrect)
+        return "Some of your classifiers are not standard classifiers:\n" + err
+
+
+class PythonClassifierVersion(BaseTest):
     def test(self, data):
         self._major_version_specified = False
 
-        classifiers = data.get('classifiers', [])
+        classifiers = data.get("classifiers", [])
         for classifier in classifiers:
-            parts = [p.strip() for p in classifier.split('::')]
-            if parts[0] == 'Programming Language' and parts[1] == 'Python':
+            parts = [p.strip() for p in classifier.split("::")]
+            if parts[0] == "Programming Language" and parts[1] == "Python":
                 if len(parts) == 2:
                     # Specified Python, but no version.
                     continue
@@ -219,92 +267,132 @@ class PythonVersion(BaseTest):
 
     def message(self):
         if self._major_version_specified:
-            return "The classifiers should specify what minor versions of "\
-                   "Python you support as well as what major version."
-        return "You should specify what Python versions you support."
+            return (
+                "The classifiers should specify what minor versions of "
+                "Python you support as well as what major version."
+            )
+        return "The classifiers should specify what Python versions you support."
+
+
+class PythonRequiresVersion(BaseTest):
+    weight = 100
+
+    def test(self, data):
+        # https://github.com/regebro/pyroma/pull/83#discussion_r955611236
+        python_requires = data.get("python_requires", None)
+
+        if not python_requires:
+            return False
+
+        try:
+            SpecifierSet(python_requires)
+        except InvalidSpecifier:
+            return False
+
+        return True
+
+    def message(self):
+        return (
+            "You should specify what Python versions you support with "
+            "the 'requires-python'/'python_requires' metadata."
+        )
 
 
 class Keywords(FieldTest):
     weight = 20
-    field = 'keywords'
+    field = "keywords"
 
 
 class Author(FieldTest):
     weight = 100
-    field = 'author'
+    field = "author"
+
+    def test(self, data):
+        """Check if author_email field contains author name."""
+        email = data.get("author_email")
+        # Pass if author name in email, e.g. "Author Name <author@example.com>"
+        return True if email and "<" in email else super().test(data)
 
 
 class AuthorEmail(FieldTest):
     weight = 100
-    field = 'author_email'
+    field = "author_email"
 
 
-class Url(FieldTest):
-    weight = 20
-    field = 'url'
-
-
-class License(FieldTest):
-    weight = 50
-    field = 'license'
-
-
-class LicenceClassifier(BaseTest):
+class Url(BaseTest):
     weight = 20
 
     def test(self, data):
-        classifiers = data.get('classifiers', [])
+        return bool(data.get("url")) or bool(data.get("project_urls"))
+
+    def message(self):
+        return (
+            "Your package should have a 'url' field with a link to the "
+            "project home page, or a 'project_urls' field, with a "
+            "dictionary of links, or both."
+        )
+
+
+class Licensing(BaseTest):
+    weight = 50
+
+    def test(self, data):
+        license = data.get("license")
+        classifiers = data.get("classifiers", [])
+        licenses = set()
         for classifier in classifiers:
-            parts = [p.strip() for p in classifier.split('::')]
-            if parts[0] == 'License':
+            parts = [p.strip() for p in classifier.split("::")]
+            if parts[0] == "License":
+                # license classifier exist
+                licenses.add(classifier)
+
+        if not license and not licenses:
+            self._message = "Your package does neither have a license field nor any license classifiers."
+            return False
+
+        if license in CODE_LICENSES:
+            if not CODE_LICENSES[license].intersection(licenses):
+                self._message = f"The license '{license}' specified is not listed in your classifiers."
+                return False
+
+        return True
+
+    def message(self):
+        return self._message
+
+
+class DevStatusClassifier(BaseTest):
+    weight = 20
+
+    def test(self, data):
+        classifiers = data.get("classifiers", [])
+        for classifier in classifiers:
+            parts = [p.strip() for p in classifier.split("::")]
+            if parts[0] == "Development Status":
                 # license classifier exist
                 return True
         return False
 
     def message(self):
-        return "You should specify license in classifiers."
-
-
-class ZipSafe(BaseTest):
-
-    def test(self, data):
-        if data.get('_setuptools'):
-            self.weight = 20
-            return 'zip_safe' in data
-        else:
-            self.weight = 0
-            return True
-
-    def message(self):
-        return "You are using Setuptools or Distribute but do not specify if "\
-               "this package is zip_safe or not. You should specify it, as "\
-               "it defaults to True, which you probably do not want."
+        return "Specifying a development status in the classifiers gives users a hint of how stable your software is."
 
 
 class SDist(BaseTest):
     weight = 100
 
     def test(self, data):
-        if '_has_sdist' not in data:
+        if "_has_sdist" not in data:
             # We aren't checking on PyPI
             self.weight = 0
             return None
-        return data['_has_sdist']
+        return data["_has_sdist"]
 
     def message(self):
-        return ("You have no source distribution on the Cheeseshop. "
-                "Uploading the source distribution to the Cheeseshop ensures "
-                "maximum availability of your package.")
-
-
-class PackageDocs(BaseTest):
-    weight = 0  # Just a recommendation
-
-    def test(self, data):
-        return data.get('_readthe_docs')
-
-    def message(self):
-        return "You might want to host your documentation on readthedocs.org."
+        return (
+            "You have no source distribution on the Cheeseshop. "
+            "Uploading the source distribution to the Cheeseshop ensures "
+            "maximum availability of your package."
+        )
 
 
 class ValidREST(BaseTest):
@@ -312,32 +400,43 @@ class ValidREST(BaseTest):
     weight = 50
 
     def test(self, data):
-        source = data.get('long_description', '')
-        try:
-            # Try to publish to HTML and see if we get an error or not.
-            publish_parts(source=source, writer_name='html4css1')
-        except SystemMessage as e:
-            self._message = e.args[0].strip()
-            return False
+        content_type = data.get("long_description_content_type", None)
+        if content_type in ("text/plain", "text/markdown"):
+            # These can't fail. Markdown will just assume everything
+            # it doesn't understand is plain text.
+            return True
 
-        return True
+        # This should be ReStructuredText
+        source = data.get("long_description", "")
+        stream = io.StringIO()
+        settings = {"warning_stream": stream}
+
+        try:
+            publish_parts(source=source, writer_name="html4css1", settings_overrides=settings)
+        except SystemMessage as e:
+            self._message = e.args[0]
+        errors = stream.getvalue().strip()
+        if not errors:
+            return True
+
+        self._message = "\n" + errors
+        return False
 
     def message(self):
-        return 'Your long_description is not valid ReST: ' + self._message
+        return "Your long_description is not valid ReST: " + self._message
 
 
 class BusFactor(BaseTest):
-
     def test(self, data):
-        if '_owners' not in data:
+        if "_owners" not in data:
             self.weight = 0
             return None
 
-        if len(data.get('_owners', [])) == 1:
+        if len(data.get("_owners", [])) == 1:
             self.weight = 100
             return False
 
-        if len(data.get('_owners', [])) == 2:
+        if len(data.get("_owners", [])) == 2:
             self.weight = 50
             return False
 
@@ -349,6 +448,41 @@ class BusFactor(BaseTest):
         return "You should have three or more owners of the project on PyPI."
 
 
+class MissingBuildSystem(BaseTest):
+    def test(self, data):
+        if "_missing_build_system" in data:
+            # These sort of "negative only/deprecation" ratings only give you negative weight
+            self.weight = 200
+            return False
+
+    def message(self):
+        return (
+            "You seem to have a setup.cfg, but neither a setup.py, nor a build-system defined. This makes "
+            "it unclear how your project should be built. You probably want to have a pyproject.toml file "
+            "with the following configuration:\n\n"
+            "    [build-system]\n"
+            '    requires = ["setuptools>=42"]\n'
+            '    build-backend = "setuptools.build_meta"\n\n'
+            'In the future this will become a hard failure and your package will be rated as "not cheese".'
+        )
+
+
+class StoneAgeSetupPy(BaseTest):
+    def test(self, data):
+        if "_stoneage_setuppy" in data:
+            # These sort of "negative only/deprecation" ratings only give you negative weight
+            self.weight = 200
+            return False
+
+    def message(self):
+        return (
+            "Your Cheese may have spoiled!! The only way to gather metadata from your package was to execute "
+            "a patched setup.py. This indicates that your package is using very old packaging techniques, "
+            "(or that your setup.py isn't executable at all), and Pyroma will soon regard that as a complete "
+            "failure!\nPlease modernize your packaging! If it is already modern, this is a bug."
+        )
+
+
 ALL_TESTS = [
     Name(),
     Version(),
@@ -357,18 +491,20 @@ ALL_TESTS = [
     Description(),
     LongDescription(),
     Classifiers(),
-    PythonVersion(),
+    ClassifierVerification(),
+    PythonClassifierVersion(),
+    PythonRequiresVersion(),
     Keywords(),
     Author(),
     AuthorEmail(),
     Url(),
-    License(),
-    LicenceClassifier(),
-    ZipSafe(),
+    Licensing(),
     SDist(),
-    PackageDocs(),
     ValidREST(),
     BusFactor(),
+    DevStatusClassifier(),
+    MissingBuildSystem(),
+    StoneAgeSetupPy(),
 ]
 
 
@@ -398,4 +534,4 @@ def rate(data):
     # Multiply good by 9, and add 1 to get a rating between
     # 1: All non-fatal tests failed.
     # 10: All tests succeeded.
-    return (good*9)//(good+bad)+1, fails
+    return (good * 9) // (good + bad) + 1, fails
